/* Part of MuttonChop
 * copyright 2012 Jezra Lickter
 * Licensed GPLv3
 */
using GLib;
using Json;
using Gst;
using Gdk;

const uint16 PORT = 2876;

enum PlayType {
  VIDEO,
  AUDIO,
  CAST,
  STREAM
}

public errordomain ControllerError {
		NO_CONFIG
}


class Controller : GLib.Object {
  public Configuration conf;
  public UriPlayer player;
  public Catcher catcher;
  private Regex regex_path_dots;
  private PlayType current_playtype;
  private MainLoop mainloop;
  private DisplayWindow window;
  private ulong window_xid;
  private string last_audio_file_path;
  public Data data;
  private bool audio_play_random;
  private bool audio_updating;
  
  //we will send events
  public signal void event( string event, string data);
  
  public Controller() throws ControllerError {
    //compile the regex'
    try{
			regex_path_dots = new Regex("\\/*\\.\\.\\/*");
    } catch (Error e) {
			stderr.printf(e.message+"\n");
		}
		
    conf = new Configuration();
		//does the conf exist?
		if (!conf.conf_exists) {
			/* set up defaults */
			//where is the public dir?
			string pd = GLib.Path.build_filename( Config.PACKAGE_DATA_DIR,"public");
			conf.insert("web_server_public_dir",  pd );
			//where are the video, casts, and audio dir?
			conf.insert("video_dir", "~/Video");
			conf.insert("audio_dir", "~/Music");
			conf.insert("cast_dir", "~/Casts");	
    }
		
    // should audio play randomly?
    audio_play_random = bool.parse( conf.get_value("audio_play_random") );
    //set the default for audio updating
    audio_updating = false;
    
    //make the database doohicky
    data = new Data();
		 
		//make the window
		window = new DisplayWindow();
		 
    //make the player
    player = new UriPlayer();
    //connect the prepare window signal
    player.prepare_video_window.connect( cb_prepare_video_window );
    player.finished_playing.connect( cb_finished_playing );
    player.playing_stopped.connect( cb_playing_stopped );
    
    //handle the updates
    player.update.connect( (p, o, v) => { 
			event(o,v); 
		});
    
    // are we setting the default volume of the player?
    string s_vol =  conf.get_value("start_up_volume");
    int vol;
    if (s_vol == "") {
			vol = 50;
		} else {
			vol = int.parse(s_vol);
		}
    player.volume(vol);
    
    //create the catcher
    catcher = new Catcher( data, conf.get_value("cast_dir") );
    
    //clean out the catcher's files
    catcher.clean( conf.get_int_value("days_to_save_played_casts") );
    
    //create the main loop
    mainloop = new MainLoop();
  }
  
  public Response player_status() {
		Response response = Response();
		size_t size;
		var status = player.status();
		//TODO: move playtype to a new method for getting system info
		status.play_type = current_playtype.to_string(); 
		
		//TODO: move audio_play_random to a new method for getting system info
		status.audio_play_random = audio_play_random;
		
		//TODO: move audio_updating to a new method for getting system info
		status.audio_updating = audio_updating;
		status.last_audio_file_path = last_audio_file_path;
		response.text = Json.gobject_to_data(status, out size);
		response.content_type = "application/json";
		response.status_code = StatusCode.OK;
		return response;
	}
	
	public int player_volume( int vol ) {
		int new_vol = player.volume( vol ); 
		event("volume", @"$new_vol");
		return new_vol;
	}
	
	public void set_audio_random(string r) {
		audio_play_random = bool.parse(r);
		event("audiorandom", r);
	}
		
  //remove ../ from the string
  private string sanitize_path(string? s) {
		if (s==null){
			return "";
		}
		try {
     return regex_path_dots.replace(s, s.length, 0,"");
	 }catch(Error e) {
		 stderr.printf(e.message+"\n");
		 return s;
	 }
  }
  
  private string build_path(string? r, string? dp, string? d, string? f) {
    dp = dp ?? "";
    d = d ?? "";
    f = f ?? "";
    r = r ?? "";
    return GLib.Path.build_filename(r, dp, d, f);
  }
   
   
  public DirectoryInfo video_directory_info(string? p){
		string sp = sanitize_path(p);
		DirectoryInfo di = new DirectoryInfo( conf.get_value("video_dir"), sp);
		return di;
	}
	
	public bool video_play( string f) {

		//what should the full base_dir to the file be?
		string filepath = build_path(conf.get_value("video_dir"), null, null, f);
		// does the file exist?
		if (FileUtils.test(filepath,GLib.FileTest.IS_REGULAR) ) {
			player.set_uri( "file://"+filepath );
			player.play();
			//set the current playtype
			set_current_playtype( PlayType.VIDEO );
			return true;
		} else {
			return false;
		}
	}
  
  private void set_current_playtype(PlayType pt) {
		if (current_playtype != pt ) {
			current_playtype = pt;
			event("playtypechange", pt.to_string() );
		}
	}
    
  public AudioList audio_list(string? artist=null) {
		var list = data.get_audio_list(artist);
		return list;
	}
	
	public AudioList audio_search(string? search_string=null) {
		return data.get_audio_search(search_string);
	}
    
  public void handle_previous() {
    switch (current_playtype) {
      case PlayType.VIDEO :
        player.jump_percent(-10);
        break;
      case PlayType.AUDIO :
				play_audio_previous();
				break;
      default:
				player.jump_percent(-10);
        break;
    }
  }
  
  public void handle_next() {
    switch (current_playtype) {
			case PlayType.VIDEO :
        player.jump_percent(10);
        break;
      case PlayType.AUDIO :
        play_audio();
        break;
      default:
				player.jump_percent(10);
        break;
    }
  }
  
  public void handle_forward() {
		player.jump_percent(1);
	}
	
	public void handle_reverse() {
		player.jump_percent(-1);
	}
  
  private void cb_playing_stopped() {
		event("playing_stopped", "");
		window.unfullscreen();
		window.hide();
	}
  
  private void cb_finished_playing() {
		event("playing_finished", "");
		cb_playing_stopped();
		// are we currently playing audio?
		if (current_playtype == PlayType.AUDIO ) {
			play_audio();
		}
	}
  
  private void cb_prepare_video_window(UriPlayer p, Element e) {
		window.bigify();
		//window.get_window().cursor = blank_cursor;
		window_xid = (ulong)Gdk.X11Window.get_xid(window.get_window());
		var xoverlay = e as XOverlay;
    xoverlay.set_xwindow_id (window_xid);
	}
  
  public void audio_update_file_data() {
		//create a tagreader
		TagReader tr = new TagReader(data);
		tr.track_info.connect( cb_audio_file_data );
		tr.started.connect( ()=>{ 
			audio_updating=true; 
			event("audio_update", "start");
		});
		tr.finished.connect( ()=>{ 
			
			/* check for removed files */
			AudioList al = data.get_audio_list();
			foreach (AudioFileInfo afi in al.get_files() ) {
				string path = afi.path;
				int id = afi.id;
				//does the file exist?
				if (!file_exists(path) ) {
					//remove the item from the database
					data.remove_audio(id);
				}
			}
			audio_updating=false; 
			event("audio_update", "finish");
			//delete the tag reader
		});
		tr.walk( conf.get_value("audio_dir") );
	}
  
  private async void cb_audio_file_data(string path, string? artist, string title, string album, int track_number) {
		//update the data
		last_audio_file_path = path;
		data.update_audio_file_data(path, artist, title, album, track_number);
		//emit this event
		event("tagread", path);
	}
		
	public void play_audio_for_id(string? id=null) {
		string path = "";
		AudioFileInfo af = new AudioFileInfo();
		if(id!=null) {
			af = data.get_audio_file_for_id(id);
		} else {
			if ( audio_play_random ) {
				af = data.get_random_audio_file();
			} else {
				af = data.get_next_audio_file();
			}
		}
		path = af.path;
		//play the path
		player.set_uri( "file://"+path );
		player.play();
		//be sure to set the play type
		set_current_playtype(PlayType.AUDIO);
	}
		
  public bool play_audio() {
		play_audio_for_id();
		return false;
	}
	
  public bool play_audio_previous() {
		AudioFileInfo afi = data.get_previous_audio_file();
		play_audio_for_id(afi.id.to_string());
		return false;
	}
  
  public void run() {
		//if a database was created, this could be the first run
		if (data.create_db) {
			//if there is an audio directory, we should rock a tag reader
			if( conf.get_value("audio_dir") != null ) {
				audio_update_file_data();
			}
		}
		//are going to auto_play?
		if (bool.parse( conf.get_value("audio_auto_play") )) {
			//are we creating the db?
			if (data.create_db) {
				//start playing in 5 seconds
				Timeout.add(1000 * 5, play_audio);
			} else {
				play_audio();
			}
		}
    mainloop.run();
  }
  
  /* methods for interacting with the catcher */
  
  public string catcher_add( string uri ) {
		EpisodeList episode_list = catcher.add_feed( uri );
		return episode_list.to_json();
	}
	
	public string catcher_get_update( string id ) {
		EpisodeList episode_list;
		if( id != "all") { 
			episode_list = catcher.get_update( id );
		} else {
			episode_list = catcher.get_updates();
		}
		return episode_list.to_json();
	}
	
	public string catcher_get_episodes(string? id) {
		size_t length;
		Episode[] episodes = catcher.get_episodes(id);
		Json.Builder b = new Json.Builder();
		b.begin_array();
		foreach(Episode episode in episodes) {
			b.begin_object();
				b.set_member_name("id");
				b.add_string_value(episode.id);
				b.set_member_name("title");
				b.add_string_value(episode.title);
				b.set_member_name("description");
				b.add_string_value(episode.description);
				b.set_member_name("pubdate");
				b.add_string_value(episode.pubdate);
				b.set_member_name("played");
				b.add_boolean_value(episode.played);
				b.set_member_name("downloaded");
				b.add_boolean_value(episode.downloaded);
				b.set_member_name("exists");
				b.add_boolean_value(file_exists(episode.local_path));
			b.end_object();
		}		
		b.end_array();
		//b.end_object();
		Generator g = new Generator();
		g.indent = 2;
		g.pretty = true;
		g.set_root( b.get_root() );
		return g.to_data( out length );
	}
	
	public string catcher_get_episode(string? id) {
		size_t length;
		Episode episode = catcher.get_episode(id);
		Json.Builder b = new Json.Builder();

		b.begin_object();
		b.set_member_name("id");
		b.add_string_value(episode.id);
		b.set_member_name("title");
		b.add_string_value(episode.title);
		b.set_member_name("description");
		b.add_string_value(episode.description);
		b.set_member_name("pubdate");
		b.add_string_value(episode.pubdate);
		b.set_member_name("played");
		b.add_boolean_value(episode.played);
		b.set_member_name("exists");
		b.add_boolean_value(file_exists(episode.local_path));
		b.set_member_name("downloaded");
		b.add_boolean_value(episode.downloaded);
		b.end_object();

		Generator g = new Generator();
		g.indent = 2;
		g.pretty = true;
		g.set_root( b.get_root() );
		return g.to_data( out length );
	}
	
	
	public void catcher_download(string id_string) {
		string[] ids = id_string.split(":");
		foreach (string id in ids ) {
			catcher.download_episode(id);
		}
	}
	
	public void catcher_kill( string id ) {
		catcher.kill_download(id);
	}
		
	public string catcher_get_feeds() {
		size_t length;
		Feed[] feeds = catcher.get_feeds();
		//json this feeds array
		Json.Builder b = new Json.Builder();
		//b.begin_object();
		//b.set_member_name("downloads");
		b.begin_array();
		foreach(Feed feed in feeds) {
			b.begin_object();
				b.set_member_name("id");
				b.add_string_value(feed.id);
				b.set_member_name("source");
				b.add_string_value(feed.source);
				b.set_member_name("title");
				b.add_string_value(feed.title);
				b.set_member_name("description");
				b.add_string_value(feed.description);
			b.end_object();
		}		
		b.end_array();
		//b.end_object();
		Generator g = new Generator();
		g.indent = 2;
		g.pretty = true;
		g.set_root( b.get_root() );
		return g.to_data( out length );
	}
	
	public void catcher_play(string id) {
		//get the path from the db
		Episode e = data.get_episode(id);
		//play the local path of the episode
		string path = e.local_path;
		//play the path
		player.set_uri( "file://"+path );
		player.play();
		//set the file as played
		data.set_episode_played( id );
		//be sure to set the play type
		set_current_playtype(PlayType.CAST);
	}
	
	public Response catcher_delete( Request req ) {
		//start building the response code
		Response res = Response();
		res.status_code = StatusCode.OK;
		res.content_type = "text/plain";
		
		string id = req.val;
		if (id == null) {
			res.status_code = StatusCode.ERROR;
			res.text = "No Feed ID was supplied";
		} else {
			catcher.remove_feed( id );
			res.text = "OK";
		}
		return res;
	}
	
	
	public string catcher_download_status() {
		DownloaderInfo[] dlis = catcher.get_download_status();
		uint64 percent;
		//create a json string for the dli information
		Json.Builder b = new Json.Builder();
		b.begin_object();
		b.set_member_name("downloads");
		b.begin_array();
		foreach(DownloaderInfo dli in dlis) {
			b.begin_object();
				b.set_member_name("id");
				b.add_string_value(dli.id);
				b.set_member_name("title");
				b.add_string_value(dli.title);
				b.set_member_name("feed-title");
				b.add_string_value(dli.feed_title);
				b.set_member_name("length");
				b.add_double_value(dli.length);
				b.set_member_name("bytes-downloaded");
				b.add_double_value(dli.bytes_downloaded);
				b.set_member_name("percent");
				if( dli.bytes_downloaded > 0 && dli.length > 0) {
					percent = dli.bytes_downloaded*100/dli.length;
				} else {
					percent = 0;
				}
				b.add_double_value(percent);
			b.end_object();
		}		
		b.end_array();
		b.end_object();
		Generator g = new Generator();
		g.indent = 2;
		g.pretty = true;
		g.set_root( b.get_root() );
		size_t len;
		return g.to_data( out len );
		
	}
	
	private bool file_exists(string? path) {
		if (path==null) {
			return false;
		}
		return FileUtils.test(path, FileTest.EXISTS) ;
	}
	
	/* stream actions */ 
	public Response stream_play( Request req ) {
			Response res = Response();
			string id = req.val;
			string src = "";
			if ( int.parse(id) > 0 ) {
				//get the source from the db
				src = data.stream_get_source_for_id( id );
			} else {
				//the "id" is the src?
				src = id;
			}
			try {
				player.set_uri( src );
				player.play();
				res.status_code = StatusCode.OK;
				res.content_type = "text/plain";
				res.text = "OK";
				//change the stream play state
				set_current_playtype( PlayType.STREAM );
			} catch (Error e) {
				res.status_code = StatusCode.ERROR;
				res.content_type = "text/plain";
				res.text = e.message;
			}
			return res;
				
	}
	public Response stream_add( Request req ) {
		Response res = Response();
		string source = req.val;
		if (source == null || source == "") {
			res.content_type = "text/plain";
			res.status_code = StatusCode.ERROR;
			res.text = "Stream URL is empty";
			return res;
		}
		string title = req.args["title"] ?? "";
		string id = data.stream_add( source, title);
		
		
		res.content_type = "text/plain";
		res.text = id;
		if (int.parse(id) < 1 ) {
			res.status_code = StatusCode.ERROR;
		} else {
			res.status_code = StatusCode.OK;
		}
		return res;
	}
	public Response stream_edit( Request req ) {
		Response res = Response();
		string id = req.val;
		string? title = req.args["title"];
		string? source = req.args["source"];
		res.text = data.stream_update(id, title, source );
		if (int.parse(res.text) < 1 ) {
			//something is wrong
			res.status_code = StatusCode.ERROR;
		} else {
			res.status_code = StatusCode.OK;
		}
		res.content_type = "text/plain";
		return res;
	}
	public Response stream_delete( Request req ) {
		Response res = Response();
		string id = req.val;
		res.text = data.stream_delete( id );
		if (int.parse(res.text) < 1 ) {
			//something is wrong
			res.status_code = StatusCode.ERROR;
		} else {
			res.status_code = StatusCode.OK;
		}
		res.content_type = "text/plain";
		return res;
	}
	public Response stream_list() {
		Response res = Response();
		res.content_type = "text/json";
		res.status_code = StatusCode.OK;
		var sl = data.stream_list();
		res.text = sl.json();
		return res;
	}
	public Response stream_info(Request req) {
		Response res = Response();
		res.content_type = "text/json";
		res.status_code = StatusCode.OK;
		var si = data.stream_info(req.val);
		size_t size;
		res.text = Json.gobject_to_data(si, out size);
		return res;
	}
	
}

