/* Part of MuttonChop
 * copyright 2012 Jezra Lickter
 * Licensed GPLv3
 */
using Gst;
using GLib;

/* class: PositionDuration 
 * usage: fill class with data and pass it along
 */
public class PositionDuration : GLib.Object {
  public int position{get; set;}
  public int duration{get; set;}
}

/* class: PlayerStatus 
 * usage: fill class with data and pass it along
 */
public class PlayerStatus :GLib.Object {
  public string state{get; set;}
  public int volume{get; set;}
  public string artist{get; set;}
  public string title{get; set;}
  public string album{get; set;}
  public int track_number{get; set;}
  public int position{get; set;}
  public int duration{get; set;}
  public string filename{get; set;}
  public string play_type{get; set;}
  public bool audio_play_random{get; set;}
  public bool audio_updating {get; set;}
  public string last_audio_file_path {get; set;}
  
  public PlayerStatus() {
    clear();
  }
  
  public void clear() {
    state="";
    volume=0;
    artist="";
    title="";
    album="";
    track_number=0;
    position=0;
    duration=0;
    filename="";
  }
  
}

public class UriPlayer : GLib.Object {
	//what signals do we have
  public signal void playing_stopped();
	public signal void finished_playing();
	public signal void playing_error();
  public signal void prepare_video_window(Gst.Element videosink);
  public signal void volume_changed(double vol);
  
	public signal void update( string obj, string val );

	//what variable need to be scoped to this class?
	private Element playbin;
	private Gst.Bus bus;
	private Query duration_query;
	private Query position_query;
	private int64 position;
	private int64 duration;
	private string track_uri;
	private PlayerStatus player_status;

	//start making the player
	public UriPlayer() {
		//set some default values

		duration=0;
		position=0;
		playbin = Gst.ElementFactory.make("playbin2","my-playbin");
		bus = playbin.get_bus();
		bus.add_signal_watch();
		bus.message.connect( bus_message );
		//define the queries
		duration_query = new Query.duration(Format.TIME);
		position_query = new Query.position(Format.TIME);
		//create the PlayerStatus
		player_status = new PlayerStatus();
		Timeout.add(1000, timed_positionduration );
	}
	
	public void set_uri(string uri) {
		//stop the playbin
		playbin.set_state(Gst.State.READY);
		//take a bit of a nap
		Thread.usleep(500000);
    //clear the collected data
    player_status.clear();
		track_uri = uri;
		player_status.filename = GLib.Path.get_basename( uri );
		//emit an update 
		update("newfile", player_status.filename);
		//load the new uri
		playbin.set_property("uri", uri);		
		//we don't want to play yet
		playbin.set_state(Gst.State.PAUSED);		
		//clear the positionduration
		update("positionduration","0:0");
		
	}
	
	~UriPlayer() {
		playbin.set_state(Gst.State.NULL);
	}

	public void handle_tags(Gst.TagList list, string tag) {
		if(tag!=null) {
			string val="Unknown";
			switch (tag) {
				case "artist":
					list.get_string(tag, out val);
					player_status.artist = val;
					update("artist", val);
					break;
				case "title":
					list.get_string(tag, out val);
					player_status.title = val;
					update("title", val);
					break;
				case "album":
          list.get_string(tag, out val);
					player_status.album = val;
					update("album", val);
          break;
				case "channel-mode":
				case "comment":
				case "audio-codec":
					break;
				case "track-number":
          uint tn=0;
          list.get_uint(tag, out tn);
					player_status.track_number = (int)tn;
          break;
				case "bitrate":
					break;
				default:
					break;
			}
		}
	}
  
	private void bus_message(Gst.Message message) {
	MessageType t = (MessageType)message.type;
	switch(t) {
		case MessageType.ERROR:
			//null the pipeline
			playbin.set_state(Gst.State.NULL);
			this.playing_error();
      stderr.printf("Error playing: %s\n", track_uri);
      //send a signal that playing is finished
      this.finished_playing();
			break;
		case MessageType.EOS:
			//null the pipeline
			playbin.set_state(Gst.State.NULL);
			//send a signal that playing is finished
			this.finished_playing();
			break;
		case	MessageType.TAG:
			Gst.TagList tag_list;
			message.parse_tag(out tag_list);
			//we need to get the key and value from the tag
			//tag_list.foreach( (l,t)=>{ handle_tags(l,t); }	);
			tag_list.foreach( handle_tags );
			break;
		case MessageType.STATE_CHANGED:
			State state;
			State pending;
			playbin.get_state(out state,out pending, 1);
			player_status.state = state.to_string();
			break;
		case MessageType.ELEMENT:
			var st = message.get_structure();
			if (st.has_name("prepare-xwindow-id") ) {
        message.src.set("force-aspect-ratio", true);
        //message.src.set("draw-borders", true);
        //emit a signal regarding this message src
        prepare_video_window((Element)message.src);
			}
      
			break;
		default:
			//what is the message type?
			break;
		}
	}

	public void stop() {
		playbin.set_state(Gst.State.READY);
    playing_stopped();
	}
	
	//when the play command is sent
	public void play() {
		playbin.set_state(Gst.State.PLAYING);
		update("playerstate", "PLAYING");
	}
	
	public void pause() {
		playbin.set_state(Gst.State.PAUSED);
		update("playerstate", "PAUSED");
	}
	
  //set or get the volume
  public int volume(int? val) {
		double vol = 0;
		if (val != null && val>=0 && val<=100 ) {
			vol = (val*1.0)/100;
      playbin.set("volume",vol);
		}
		playbin.get("volume", out vol);
		val = (int)(vol*100);
		return val;
	}
  
   public PlayerStatus status() {
    //we will be returning a player status
		//what is the position and duration?
		PositionDuration pd = position_duration();
		player_status.position = pd.position;
		player_status.duration = pd.duration;
		//get the volume
		player_status.volume = volume(null);
    return player_status;
  }
  
  //jump jp percent relative to current playback
  public void jump_percent( int jp ) {
		//get the current position/duration
		PositionDuration pd = position_duration();
		//what will the new offset be?
		int offset = (pd.duration/100)*jp;
		//the new position is postion + offset
		int64 newpos = (pd.position+offset);
		if (0 <= newpos < pd.duration) {
			//seek to the new position
			newpos*=1000000000;
			playbin.seek_simple(Format.TIME,SeekFlags.KEY_UNIT|SeekFlags.FLUSH, newpos);
		}
	}
  
  //move playback to an absolute percentile position
  public PositionDuration absolute_percent(double? val) {
    if (val != null && val>=0 && val<=100 ) {
      PositionDuration pd = position_duration();
      int64 newpos = (int64)(pd.duration*val)/100;
      //seek to the new position
			newpos*=1000000000;
			playbin.seek_simple(Format.TIME,SeekFlags.KEY_UNIT|SeekFlags.FLUSH, newpos);
    }
    return position_duration();
  }
  
	//we will need to get the current position and duration of the playbin 
	public PositionDuration position_duration() {
    PositionDuration pd = new PositionDuration();
    pd.position = 0;
    pd.duration = 0; 
    bool duration_result,position_result;
		Format format = Format.TIME;
		duration_result = playbin.query(this.duration_query);
		position_result = playbin.query(this.position_query);
		if ( duration_result && position_result	) {
			this.duration_query.parse_duration(out format, out this.duration);
			this.position_query.parse_position(out format, out this.position);
		}
		// oh man, we don't need to be so accurate. using seconds should work
		pd.position = (int)(this.position/1000000000);
		pd.duration = (int)(this.duration/1000000000);

    return(pd);
  }
  
  private bool timed_positionduration() {
		event_positionduration();
		return true;
	}
	
	private void event_positionduration() {
		State state;
		State pending;
		playbin.get_state(out state,out pending, 1);
		if (player_status.state == "PLAYING") {
			var pd = position_duration();
			update("positionduration", @"%d:%d".printf(pd.position,pd.duration) );
		}
	}
}
